;***************************************************************************;
;                           K B D I N T . A S M                             ;
;===========================================================================;
;       Keyboard interrupt experiment.                                      ;
;       Assemble with Turbo Assembler version 3.0 or later.                 ;
;       Link with Turbo Linker version 5.1 or later.                        ;
;---------------------------------------------------------------------------;
;       Copyright (c) Simon Groes, 1994                                     ;
;***************************************************************************;

        IDEAL
        DOSSEG
        MODEL   small
        STACK   256
        LOCALS

;=====  INCLUDES ===============================================
        INCLUDE "common.inc"
        INCLUDE "macro.inc"
;---------------------------------------------------------------

;=====  EQUATES ================================================
INT09_RIGHT     =       0
INT09_WRONG1    =       1
INT09_WRONG2    =       2
INT09_ROUTINE   =       INT09_RIGHT
;---------------------------------------------------------------

        DATASEG

bExitCode       db      0       ; For terminating program.

bKbhit          db      0       ; If key hit = 1, else = 0.

szResult0       db      CR,LF,' ''bKbhit'' variable was incremented ',0
szResult1       db      ' time(s).',0
szPrompt        db      CR,LF,'Press any key...',0

szAddrOld       db      CR,LF,'The old interrupt address was: ',0
szAddrNew       db      CR,LF,'The new interrupt address is : ',0
szAddrRet       db      CR,LF,'The interrupt address is restored to: ',0
szNumBuf        db      7 DUP (0)

        CODESEG

;=====  EXTRN PROCEDURES =======================================
;-----  From DOSIO module
        EXTRN   DOSIO_PutString:PROC, DOSIO_PutChar:PROC
;-----  From MATHIO module
        EXTRN   MATHIO_BinToAsc:PROC
;---------------------------------------------------------------

;***************************************************************
;                     MAIN PROGRAM
;***************************************************************
Start:
        mov     ax, @data
        mov     ds, ax
        mov     es, ax

;-----  Save old interrupt address.
        push    es
        mov     ax, 3509h               ; Function 35h - Get int vector.
        int     21h                     ; es=seg, bx=ofs
        mov     [cs:wInt09Ofs], bx
        mov     [cs:wInt09Seg], es
        pop     es

;-----  Display old interrupt address.
        mov     di, OFFSET szAddrOld    ; 'The old interrupt address was:'
        call    DOSIO_PutString
        mov     ax, [cs:wInt09Seg]      ; Input for 'DisplayAddress'.
        mov     dx, [cs:wInt09Ofs]
        call    DisplayAddress          ; Call subroutine - display addr.

;-----  Set new interrupt address.
        push    ds
        mov     ax, 2509h               ; Function 25h - Set int vector.
        push    cs
        pop     ds                      ; ds = cs.
        mov     dx, OFFSET Int09
        int     21h                     ; New int vector set.
        pop     ds

;-----  Display New Interrupt Address.
        mov     di, OFFSET szAddrNew    ; 'The interrupt address now is: '
        call    DOSIO_PutString         
        mov     ax, cs                  ; Seg addr in AX.
        mov     dx, OFFSET Int09        ; Ofs addr in DX.
        call    DisplayAddress  ; Seg & ofs addr of new int 9 displayed.

;-----  Prompt user to press key.
        mov     di, OFFSET szPrompt     ; 'Press any key...'
        call    DOSIO_PutString

;-----  Test new interrupt.
        mov     cx, 444h                ; Don't allow loop to go on forever.
TestLoop:
        hlt                             ; Wait for interrupt.
        cmp     [bKbhit], 1             ; [bKbhit] = 1 ?
        loopnz  TestLoop                ; No : Continue.

;-----  Restore old interrupt vector.
        push    ds
        mov     ax, 2509h               ; Function 25h, int 09h.
        mov     dx, [cs:wInt09Ofs]
        mov     ds, [cs:wInt09Seg]
        int     21h
        pop     ds

;-----  Display result.
        mov     di, OFFSET szResult0    ; ''bKbhit' was incremented '
        call    DOSIO_PutString

        mov     al, [bKbhit]            ; Set up params for BinToAsc:PROC.
        xor     ah, ah                  ; AX = value.
        mov     bx, 10                  ; bx = base (2,10 or 16).
        mov     di, OFFSET szNumBuf     ; di -> buffer to hold ASCII digits.
        xor     cx, cx                  ; cl = min # digits, ch = type of
                                        ;  leading char ('0' or SPACE).
        call    MATHIO_BinToAsc         ; Value of 'bKbhit' as ASCII string.
        call    DOSIO_PutString         ; Value displayed.
        mov     di, OFFSET szResult1    ; ' time(s)'
        call    DOSIO_PutString

;-----  Get and display restored interrupt address.
        push    es
        mov     ax, 3509h       ; Function 35h - Get int vector.
        int     21h             ; es=seg, bx=ofs
        mov     dx, bx          ; dx = offset address, ax = segment address
        mov     ax, es          ;  for 'DisplayAddress' procedure.
        pop     es
        mov     di, OFFSET szAddrRet    ; Display restored address.
        call    DOSIO_PutString
        call    DisplayAddress

;-----  Exit program.
        mov     ah, 04Ch                ; Function 4Ch - Exit program.
        mov     al, [bExitCode]         ; Return with exit code.
        int     21h                     ; Program terminated.
;***************************************************************

;===============================================================
; Subroutine:   =*=DisplayAddress=*=
;---------------------------------------------------------------
; Task:         Display SEG:OFS address of interrupt.
; Input:        ax = segment address.
;               dx = offset address.
; Output:       none
; Registers:    none
;===============================================================
PROC    DisplayAddress

        SaveRegs <ax,bx,cx,di>
        mov     bx, 16                  ; Base 16 (hex).
        mov     cl, 4                   ; Minimum 4 digits.
        mov     ch, '0'                 ; Fill leading digits with 0.
        mov     di, OFFSET szNumBuf     ; Buffer to hold digits.
        call    MATHIO_BinToAsc
        call    DOSIO_PutString         ; 'XXXXH'displayed.
        mov     al, ':'                 
        call    DOSIO_PutChar           ; 'XXXXH:' displayed.
        mov     ax, dx                  ; ax = offset address.
        call    MATHIO_BinToAsc
        call    DOSIO_PutString         ; 'XXXXH:XXXXH' displayed.
        RestoreRegs <di,cx,bx,ax>
        ret
ENDP    DisplayAddress
;...............................................................

;-----  THE CORRECT INT 9 HANDLER ROUTINE.
; This routine allows the original handler for int 9, when it is called,
; to call itself with the 'int 9' instruction, without affecting the 
; operation of the original or new interrupt handlers for int 9.
; The new handler's processing is only executed once when the original 
; handler returns.

IF INT09_ROUTINE EQ INT09_RIGHT
;===============================================================
; Interrupt:    =*=Int09=*=
;---------------------------------------------------------------
; Task:         Increment bKbhit variable at keyboard int (09h).
; Input:        none
; Output:       none
; Registers:    none
;===============================================================

;=====  VARIABLES ==============================================
fRunning        db      0               ; To avoid reentry.

LABEL   dwInt09LPtr     DWORD
wInt09Ofs       dw      0       ; Offset address of old 1Ch int.
wInt09Seg       dw      0       ; Segment address of old 1Ch int.
;---------------------------------------------------------------
PROC    Int09

        cmp     [BYTE cs:fRunning], 0   ; Program already running ?
        jne     @@Default               ; Yes: Execute only original int 9.
        inc     [BYTE cs:fRunning]      ; No : Set [fRunning] to 1.
        
        pushf                           ; Simulate 'INT' instruction.
        call    [DWORD cs:dwInt09LPtr]  ; Call original int 9h.

;-----  New handler's processing.
        sti                             ; Allow interrupts.
        SaveRegs <ax, ds>
        
        mov     ax, @data
        mov     ds, ax                  ; ds -> local dataseg.

        inc     [BYTE ds:bKbhit]

;-----  Disable interrupts during cleanup and exit.
        cli                             ; Disable interrupts.
        dec     [BYTE cs:fRunning]      ; Reset [fRunning] flag to 0.
        RestoreRegs <ds, ax>

@@End:
        iret
@@Default:
        pushf                           ; Simulate 'INT' instruction.
        call    [DWORD cs:dwInt09LPtr]  ; Call original int 9h.
        iret
ENDP    Int09
;...............................................................
ENDIF ; INT09_ROUTINE EQ INT09_RIGHT


;-----  INCORRECT INT 9 HANDLER ROUTINE #1
; This routine assumes that the original handler never uses the 'int 9'
; instruction.

IF INT09_ROUTINE EQ INT09_WRONG1
;===============================================================
; Interrupt:    =*=Int09=*=     !! WRONG !!
;---------------------------------------------------------------
; Task:         Increment bKbhit variable at keyboard int (09h).
; Input:        none
; Output:       none
; Registers:    none
;===============================================================

;=====  VARIABLES ==============================================
fRunning        db      0               ; To avoid reentry.

LABEL   dwInt09LPtr     DWORD
wInt09Ofs       dw      0       ; Offset address of old 1Ch int.
wInt09Seg       dw      0       ; Segment address of old 1Ch int.
;---------------------------------------------------------------
PROC    Int09

        cmp     [BYTE cs:fRunning], 0   ; Program already running ?
        jne     @@End                   ; Yes: End, don't execute again.
        inc     [BYTE cs:fRunning]      ; No : Set [fRunning] to 1.
        
        pushf                           ; Simulate 'INT' instruction.
        call    [DWORD cs:dwInt09LPtr]  ; Call original int 9h.
        
        sti                             ; Allow interrupts.
        SaveRegs <ax, ds>
        
        mov     ax, @data
        mov     ds, ax                  ; ds -> local dataseg.

        inc     [BYTE ds:bKbhit]

;-----  Disable interrupts during cleanup and exit.
        cli                             ; Disable interrupts.
        dec     [BYTE cs:fRunning]      ; Reset [fRunning] flag to 0.
        RestoreRegs <ds, ax>

@@End:
        iret
ENDP    Int09
;...............................................................
ENDIF ; INT09_ROUTINE EQ INT09_WRONG2


;-----  INCORRECT INT 9 HANDLER ROUTINE #2
; This routine allows the original handler for int 9 to call itself with the 
; 'int 9'instruction, but also causes the new interrupt handler's processing 
; to be executed each time the 'int 9' instruction is used.
; The new handler's processing is executed again when the original interrupt 
; handler returns.  Results may thus be unexpected.

IF INT09_ROUTINE EQ INT09_WRONG2
;===============================================================
; Interrupt:    =*=Int09=*=     !! WRONG !!
;---------------------------------------------------------------
; Task:         Increment bKbhit variable at keyboard int (09h).
; Input:        none
; Output:       none
; Registers:    none
;===============================================================

;=====  VARIABLES ==============================================
fRunning        db      0               ; To avoid reentry.

LABEL   dwInt09LPtr     DWORD
wInt09Ofs       dw      0       ; Offset address of old 1Ch int.
wInt09Seg       dw      0       ; Segment address of old 1Ch int.
;---------------------------------------------------------------
PROC    Int09

        pushf                           ; Simulate 'INT' instruction.
        call    [DWORD cs:dwInt09LPtr]  ; Call original int 9h.

        cmp     [BYTE cs:fRunning], 0   ; Program already running ?
        jne     @@End                   ; Yes: End, don't execute again.
        inc     [BYTE cs:fRunning]      ; No : Set [fRunning] to 1.
        
        sti                             ; Allow interrupts.
        SaveRegs <ax, ds>
        
        mov     ax, @data
        mov     ds, ax                  ; ds -> local dataseg.

        inc     [BYTE ds:bKbhit]

;-----  Disable interrupts during cleanup and exit.
        cli                             ; Disable interrupts.
        dec     [BYTE cs:fRunning]      ; Reset [fRunning] flag to 0.
        RestoreRegs <ds, ax>

@@End:
        iret
ENDP    Int09
;...............................................................
ENDIF ; INT09_ROUTINE EQ INT09_WRONG1

        END     Start           ; End of program / entry point.
